@@ -8,10 +8,15 @@ use warnings;
 use XML::Simple;
 use LWP::UserAgent;
 use Carp;
-use Data::Dumper qw(Dumper);
 use Modern::Perl;
 use Params::Validate qw(:all);
 
+use Data::Dumper;
+
+with 'Firewall::PaloAlto::Common';
+use Firewall::PaloAlto::SystemInfo;
+use Firewall::PaloAlto::Routes;
+
 
 =head1 NAME
 
@@ -23,7 +28,7 @@ Version 0.03
 
 =cut
 
-our $VERSION = '0.03';
+our $VERSION = '0.04';
 
 =head1 SYNOPSIS
 
@@ -64,7 +69,9 @@ The constructor generates a new object to interact with a specific firewall.
 The host, username and password parameters are mandatory.
 If not specified, SSL is used, but it can be disabled using the argument ssl => 0
 
-Detailed debugging can be turned on using the debu => 1 argument. It is off by default.
+Detailed debugging can be turned on using the debug => 1 argument. It is off by default.
+
+This method is inherited from the Firewall::PaloAlto::Common role.
 
     my $palo = Firewall::PaloAlto->new(host => 'paloalto.local', username => 'admin', password => 'admin', ssl => 0, debug => 1);
 
@@ -91,241 +98,17 @@ debug (optional, default: 0) - print debugging messages.
 
 =cut
 
-
-has host => ( is => 'ro',
-                isa => 'Str',
-                required => '1',
-);
-
-has username => ( is => 'ro',
-                    isa => 'Str',
-                    required => '1',
-);
-
-has password => ( is => 'ro',
-                    isa => 'Str',
-                    required => '1',
-);
-
-has api_key => ( is => 'rw', isa => 'Str', init_arg => undef, writer => '_api_key');
-
-has ssl => ( is => 'ro', isa => 'Bool', default => 1 );
-has base_uri => (   is => 'ro', 
-                    isa => 'Str', 
-                    lazy => '1',
-                    init_arg => undef,
-                    default => sub { return 'http' . ($_[0]->ssl ? 's' : '') . '://' . $_[0]->host . '/api/?'; } 
-
-);
-
-has 'debug' => ( is => 'ro', isa => 'Bool', required => 0, default => 0 );
-
-sub _create_requester {
-    my $self = shift;
-    my %args = @_;
-
-    my $request =  $self->base_uri. 'key=' . $self->api_key;
-
-    for my $key (keys %args) {
-        #Add the argument on to the command line
-        $request .= "&$key=$args{$key}";
-    }
-
-
-    return sub {
-        my (%request_args) = @_;
-        my $user_agent = LWP::UserAgent->new;
-
-        $self->_debug_print("API Request action", $args{action});
-
-        for my $key (keys %request_args) {
-            $self->_debug_print("API Request String", "$key=$request_args{$key}");
-            $request .= "&$key=$request_args{$key}";
-        }
-
-        my $http_request = HTTP::Request->new(GET => $request);
-        my $http_response = $user_agent->request($http_request);
-
-        return $self->_check_response_and_parse($http_response);
-    }
-}
-
-sub _check_response_and_parse {
-    my $self = shift;
-    my $http_response = shift;
-   
-    #We locally redefine croak so we can get some nice red printing around it.
-    no warnings 'redefine';
-    local *original_croak = \&croak;
-    local *croak = sub {  original_croak("\e[31m".$_[0]."\e[0m"); };
-    
-    #Check the http response message - croak if its not successful
-    croak "*[HTTP Request] Failed: ".$http_response->content if !$http_response->is_success;
-    
-    my $palo_response = XMLin($http_response->content, KeyAttr => 'name');
-    
-    #If the response doesn't contain a code (i.e. the keygen request doesn't), then we check the status string.
-    #If its successful, we assign our own 'custom code' as a workaround
-    if (!defined $palo_response->{code} && $palo_response->{status} eq "success") {
-        $palo_response->{code} = 1023;
-    }
-    
-    #Check the response from the Palo Alto:
-    my $code_info = $self->_get_response_info($palo_response->{code});
-    my $api_error = $code_info->{parse}->($palo_response);
-
-    carp "*[API Request] $code_info->{string}: $api_error" and return 0 if $code_info->{is_error};
-    
-    $self->_debug_print("API Response", $palo_response->{status});
-    
-    return $palo_response;
-}
-
-sub _get_response_info {
-    my ($self, $code) = @_;
-
-    my $response_codes = {
-                            400 => {is_error => 1, string => "Bad Request", 
-                                parse => sub { "Error 400" } },
-                            403 => { is_error => 1, string => "Forbidden", 
-                                parse => sub { $_[0]->{result}->{msg}; } 
-                            },
-                            1 => { is_error => 1, string => "Unknown Command", 
-                                parse => sub { "Error 1" } },
-                            2 => { is_error => 1, string => "Internal Error (2)", 
-                                parse => sub { "Error 2" } },
-                            3 => { is_error => 1, string => "Internal Error (3)", 
-                                parse => sub { "Error 3" } },
-                            4 => { is_error => 1, string => "Internal Error (4)", 
-                                parse => sub { "Error 4" } },
-                            5 => { is_error => 1, string => "Internal Error (4)", 
-                                parse => sub { "Error 5" } },
-                            6 => { is_error => 1, string => "Bad XPath", 
-                                parse => sub { "Error 6" } },
-                            7 => { is_error => 1, string => "Object not present", 
-                                parse => sub { "Error 7" } },
-                            8 => { is_error => 1, string => "Object not unique", 
-                                parse => sub { "Error 8" } },
-                            9 => { is_error => 1, string => "Internal Error (9)", 
-                                parse => sub { "Error 9" } },
-                            10 => { is_error => 1, string => "Reference count not zero", 
-                                parse => sub { "Error 10" } },
-                            11 => { is_error => 1, string => "Internal Error (11)", 
-                                parse => sub { "Error 11" } },
-                            12 => { is_error => 1, string => "Invalid Object", 
-                                parse => sub { ref($_[0]->{msg}->{line}) eq 'ARRAY' ? join(" / ", @{ $_[0]->{msg}->{line} }) : $_[0]->{msg}->{line} } 
-                                },
-                            13 => { is_error => 1, string => "Operation Failed", 
-                                parse => sub { $_[0]->{msg}->{line} } },
-                            14 => { is_error => 1, string => "Operation Not Possible", 
-                                parse => sub { "Error 14" } },
-                            15 => { is_error => 1, string => "Operation Denied", 
-                                parse => sub { "Error 15" } },
-                            16 => { is_error => 1, string => "Unauthorized", 
-                                parse => sub { "Error 16" } },
-                            17 => { is_error => 1, string => "Invalid Command", 
-                                parse => sub { "Error 16" } },
-                            18 => { is_error => 1, string => "Malformed XML", 
-                                parse => sub { $_[0]->{msg}->{line} } },
-                            19 => { is_error => 0, string => "Get Request Successful", 
-                                parse => sub{} },
-                            20 => { is_error => 0, string => "Set Request Successful", 
-                                parse => sub{} },
-                            21 => { is_error => 1, string => "Internal Error (21)", 
-                                parse => sub { "Error 21" } },
-                            22 => { is_error => 1, string => "Session Timed Out", 
-                                parse => sub { "Error 22" } },
-                            #Custom code for keygen success
-                            1023 => { is_error => 0, string => "KeyGen Successful", parse => sub {} },
-                        };
-
-    #Return the hash, or 'undef' if the code doesn't exist.
-    return $response_codes->{$code}; 
-}
-
-sub _generate_elements {
-    my $self = shift;
-    my %element_hash = @_;
-    my $element_string = "";
-
-    for my $key (keys %element_hash) {
-        $element_string .= "<$key>";
-
-        #If our hash points to an array reference, we iterate through the array and add member.
-        #This creates <key><member>a</member><member>b</member></key>
-        if (ref($element_hash{$key}) eq "ARRAY") {
-            for my $member (@{ $element_hash{$key} }) {
-                   $element_string .= "<member>$member</member>";
-            }
-        }
-        #If we're pointing to another hash, we recurse down, as the structure will be the same.
-        #This allows us to do <key><another-key>data</another-key></key>
-        elsif (ref($element_hash{$key}) eq "HASH") {
-            $element_string .= $self->_generate_elements(%{ $element_hash{$key} });
-        }
-        #Otherwise its just a normal <key>value</key>
-        else {
-            $element_string .= "$element_hash{$key}";
-        }
-
-        $element_string .= "</$key>";
-    }
-
-    return $element_string;
-}
-
-sub _debug_print {
-    my $self = shift;
-    my ($category, $debug_string, $colourise_sub) = @_;
-    my $string_colour = "\e[0;36m"; 
-    my $string_norm = "\e[0m";
-
-    if (!$self->debug()) {
-        return 0;
-    }
-
-    #We pass code in $colorise_sub - if it evaluates to true, we print the category in green
-    #If its false, we print in red. If its not defined, we print in orange.
-    if (defined $colourise_sub) { 
-        $string_colour = $colourise_sub->() ? "\e[32m" : "\e[31m";
-    }
-
-    say "*[".$string_colour.$category.$string_norm."] $debug_string";
-}
-
-=head1 OBJECT METHODS 
+=head1 CONFIGURATION METHODS 
 
 =head2 connect()
 
 The connect functions connects to the Palo Alto, validates and saves the API key.
-It has no parameters
+It has no parameters and is inherited from the Firewall::PaloAlto::Common role.
 
- $pa->connect();
+    $pa->connect();
 
 =cut
 
-sub connect {
-    my $self = shift; 
-
-    my $user_agent = LWP::UserAgent->new(ssl_opts => { verify_hostname => 0 });
-    my $key_request = HTTP::Request->new(GET => $self->base_uri.'type=keygen&user='.$self->username.'&password='.$self->password);
-
-    $self->_debug_print("Key Request", "");
-
-    my $key_response = $user_agent->request($key_request);
-
-    my $pa_response = $self->_check_response_and_parse($key_response);
-
-    if ($pa_response->{status} eq 'success') {
-        $self->_api_key($pa_response->{result}->{key});
-        return 1;
-    }
-    else {
-        return 0;
-    }   
-}
-
-
 =head2 commit(%parameters)
 
 The commit function commits the current candidate configuration to the Palo Alto firewall.
@@ -783,6 +566,43 @@ sub security_rule {
     return $requester->(xpath => "/config/devices/entry/vsys/entry[\@name=\'$vsys\']/rulebase/security/rules/entry[$rule_name]", element => $elements);
 }
 
+=head1 OPERATIONAL METHODS
+
+=head2 get_system_info()
+
+Returns a HASHREF containing information about the system.
+
+=cut
+
+sub get_system_info {
+    my $self = shift;
+
+    my $requester = $self->_create_requester(type => 'op');
+    my $result = $requester->(cmd => '<show><system><info></info></system></show>');
+    return Firewall::PaloAlto::SystemInfo->new( $result );
+}
+
+=head2 get_routing_table()
+
+Retrieves the routing table of a particular router. If no virtual router is specified, defaults to the 'default' virtual router.
+
+The return value is an object of type Firewall::PaloAlto::Routes.
+
+=cut
+
+sub get_routing_table {
+    my $self = shift;
+    my %args = @_;
+
+    #Use the 'default' virtual router if none is defined.
+    $args{virtual_router} //= 'default';
+
+    my $requester = $self->_create_requester(type => 'op');
+    my $result = $requester->(cmd => "<show><routing><route><virtual-router>$args{virtual_router}</virtual-router></route></routing></show>");
+
+    return Firewall::PaloAlto::Routes->new( $result );
+}
+
 =head1 AUTHOR
 
 Greg Foletta, C<< <greg at foletta.org> >>
